from typing import Optional, Tuple

import numpy as np

from PyQt5 import QtCore
from PyQt5.QtGui import QPen, QBrush, QColor, QFont
from qwt import QwtPlot, QwtPlotCurve, QwtPlotGrid, QwtText

from ..product.mode import Voltage, Current

__all__ = ["IvcViewer"]


class IvcViewer(QwtPlot):
    min_border_voltage = 1.0
    min_border_current = 0.5

    # Is mouse over plot, Tuple[float, float] (I, V)
    mouse_moved = QtCore.pyqtSignal(bool, tuple)

    min_borders_changed = QtCore.pyqtSignal()

    reference_curve_changed = QtCore.pyqtSignal()  # Reference curve changed
    test_curve_changed = QtCore.pyqtSignal()  # Test curve changed
    curve_changed = QtCore.pyqtSignal()  # Reference or test curve changed

    def __init__(self, owner, track_mouse=False, parent=None):
        super().__init__(parent)

        self.__owner = owner

        self.__reference_curve = None
        self.__reference_curve_plot = QwtPlotCurve()
        self.__reference_curve_plot.setPen(QPen(QColor(255, 0, 0, 200), 4))
        self.__reference_curve_plot.attach(self)

        self.__test_curve = None
        self.__test_curve_plot = QwtPlotCurve()
        self.__test_curve_plot.setPen(QPen(QColor(0, 0, 0, 200), 4))
        self.__test_curve_plot.attach(self)

        self.__grid = QwtPlotGrid()
        self.__grid.enableXMin(True)
        self.__grid.enableYMin(True)
        self.__grid.setMajorPen(QPen(
            QColor(0, 0, 0), 0, QtCore.Qt.SolidLine))
        self.__grid.setMinorPen(QPen(
            QColor(128, 128, 128), 0, QtCore.Qt.DotLine))
        # self.__grid.updateScaleDiv(20, 30)
        self.__grid.attach(self)

        # WTF?! TODO: Refactor m away!
        m = 640000  # Is said to be enough for anybody
        black_pen = QPen(QColor(0, 0, 0), 2)

        AxisFont = QFont()
        AxisFont.pointSize = 20
        self.setCanvasBackground(QBrush(QColor(0xe1, 0xed, 0xeb), 1))

        # X Axis
        x_axis = QwtPlotCurve()
        x_axis.setPen(black_pen)
        x_axis.setData((-m, m), (0, 0))
        x_axis.attach(self)
        self.setAxisFont(QwtPlot.xBottom, QFont("Consolas", 20))
        self.setAxisMaxMajor(QwtPlot.xBottom, 5)
        self.setAxisMaxMinor(QwtPlot.xBottom, 5)
        t = QwtText(QtCore.QCoreApplication.translate("t", "\nНапряжение (В)"))
        t.setFont(AxisFont)
        self.setAxisTitle(QwtPlot.xBottom, t)

        # Y Axis
        y_axis = QwtPlotCurve()
        y_axis.setPen(black_pen)
        y_axis.setData((0, 0), (-m, m))
        y_axis.attach(self)
        self.setAxisMaxMajor(QwtPlot.yLeft, 5)
        self.setAxisMaxMinor(QwtPlot.yLeft, 5)
        self.setAxisFont(QwtPlot.yLeft, QFont("Consolas", 20))
        t = QwtText(QtCore.QCoreApplication.translate("t", "Ток (мА)\n"))
        t.setFont(AxisFont)
        self.setAxisTitle(QwtPlot.yLeft, t)

        # Initial setup for axis scales
        self.__min_border_voltage = abs(float(IvcViewer.min_border_voltage))
        self.__min_border_current = abs(float(IvcViewer.min_border_current))
        self.setAxisScale(
            QwtPlot.xBottom,
            -self.__min_border_voltage,
            self.__min_border_voltage)
        self.setAxisScale(
            QwtPlot.yLeft,
            -self.__min_border_current,
            self.__min_border_current)

    # min_bounds management
    def get_min_borders(self) -> Tuple[float, float]:
        return self.__min_border_voltage, self.__min_border_current

    def set_min_borders(
            self,
            voltage: Optional[float] = None,
            current: Optional[float] = None
    ) -> None:
        self.__set_min_borders(voltage=voltage, current=current)
        self.__adjust_scale()
        self.min_borders_changed.emit()

    def clear_min_borders(self) -> None:
        self.__min_border_voltage = abs(float(IvcViewer.min_border_voltage))
        self.__min_border_current = abs(float(IvcViewer.min_border_current))
        self.__adjust_scale()
        self.min_borders_changed.emit()

    min_borders = property(get_min_borders, set_min_borders, clear_min_borders)

    # reference_curve management
    def get_reference_curve(self) -> Optional[tuple]:
        return self.__reference_curve

    def set_reference_curve(self, reference_curve: Optional[tuple]):
        self.__set_reference_curve(reference_curve)
        self.__adjust_scale()
        self.reference_curve_changed.emit()
        self.curve_changed.emit()

    def clear_reference_curve(self):
        self.__set_reference_curve(None)
        self.__adjust_scale()
        self.reference_curve_changed.emit()
        self.curve_changed.emit()

    reference_curve = property(
        get_reference_curve, set_reference_curve, clear_reference_curve)

    # test_curve management
    def get_test_curve(self) -> Optional[tuple]:
        return self.__test_curve

    def set_test_curve(self, test_curve: Optional[tuple]) -> None:
        self.__set_test_curve(test_curve)
        self.__adjust_scale()
        self.test_curve_changed.emit()
        self.curve_changed.emit()

    def clear_test_curve(self):
        self.__set_test_curve(None)
        self.__adjust_scale()
        self.test_curve_changed.emit()
        self.curve_changed.emit()

    test_curve = property(get_test_curve, set_test_curve, clear_test_curve)

    # curves management
    def get_curves(self) -> Tuple[Optional[tuple], Optional[tuple]]:
        return self.__reference_curve, self.__test_curve

    def set_curves(
            self,
            curves: Optional[Tuple[Optional[tuple], Optional[tuple]]] = None,
            *args,
            reference_curve: Optional[tuple] = None,
            test_curve: Optional[tuple] = None,
            voltage_min_border: Optional[float] = None,
            current_min_border: Optional[float] = None,
            meas_current: Optional[Current] = None,
            meas_volt: Optional[Voltage] = None
    ) -> None:
        """Set curves from tuple or by names."""

        if len(args) > 0:
            raise ValueError("Single argument or named arguments required")

        self.__set_curves(curves=curves, reference_curve=reference_curve,
                          test_curve=test_curve)
        self.__set_min_borders(voltage=voltage_min_border,
                               current=current_min_border)
        self.__adjust_scale(measure_current=meas_current, measure_volt=meas_volt)
        self.reference_curve_changed.emit()
        self.test_curve_changed.emit()
        self.curve_changed.emit()

    def clear_curves(self):
        self.__set_curves(reference_curve=None, test_curve=None)
        self.__adjust_scale()
        self.reference_curve_changed.emit()
        self.test_curve_changed.emit()
        self.curve_changed.emit()

    curves = property(get_curves, set_curves, clear_curves)

    def __set_min_borders(
            self,
            voltage: Optional[float] = None,
            current: Optional[float] = None
    ) -> None:
        if voltage is not None:
            self.__min_border_voltage = abs(float(voltage))
        if current is not None:
            self.__min_border_current = abs(float(current))

    def __set_reference_curve(self, reference_curve: Optional[tuple]):
        self.__reference_curve = reference_curve
        _plot_curve(self.__reference_curve, self.__reference_curve_plot)

    def __set_test_curve(self, test_curve: Optional[tuple]):
        self.__test_curve = test_curve
        _plot_curve(self.__test_curve, self.__test_curve_plot)

    def __set_curves(
            self,
            curves: Optional[Tuple[Optional[tuple], Optional[tuple]]] = None,
            reference_curve: Optional[tuple] = None,
            test_curve: Optional[tuple] = None
    ) -> None:
        if curves is not None:
            reference_curve, test_curve = curves
        if test_curve is not None:
            self.__set_test_curve(test_curve)
        if reference_curve is not None:
            self.__set_reference_curve(reference_curve)

    ADJUST_SCALE_TABLE = {
        Voltage.VOLTAGE_3_3V: {
            Current.CURRENT_250UA: (4.0, 0.15),
            Current.CURRENT_2_5MA: (4.0, 1.0),
            Current.CURRENT_25MA: (4.0, 10.0),
        },
        Voltage.VOLTAGE_5V: {
            Current.CURRENT_250UA: (6.0, 0.18),
            Current.CURRENT_2_5MA: (6.0, 1.5),
            Current.CURRENT_25MA: (6.0, 15.0),
        },
        Voltage.VOLTAGE_12V: {
            Current.CURRENT_250UA: (14.0, 0.35),
            Current.CURRENT_2_5MA: (14.0, 2.8),
            Current.CURRENT_25MA: (14.0, 28.0),
        }
    }

    def __adjust_scale(
            self,
            measure_current: Optional[Current] = None,
            measure_volt: Optional[Voltage] = None
    ) -> None:
        voltage = measure_volt
        current = measure_current

        if voltage is None:
            voltage = self.__owner.voltage
        if current is None:
            current = self.__owner.current

        border_voltage, border_current =\
            self.ADJUST_SCALE_TABLE[voltage][current]
        self.setAxisScale(QwtPlot.xBottom, -border_voltage, border_voltage)
        self.setAxisScale(QwtPlot.yLeft, -border_current, border_current)


def _plot_curve(curve: Optional[tuple], curve_plot: QwtPlotCurve) -> None:
    if curve is None or curve == (None, None):
        curve_plot.setData((), ())
    else:
        # Get curves and close the loop
        voltages = np.append(curve[1][0], curve[1][0][0])
        currents = np.append(curve[1][1], curve[1][1][0]) * 1000

        # Setting curve data: (voltage [V], current [mA])
        curve_plot.setData(voltages, currents)
