import os
import os.path
import enum
import math
from typing import Optional, Set, Tuple

import numpy as np
from PyQt5 import QtCore
from PyQt5.QtGui import QPixmap
from PyQt5.QtWidgets import QMainWindow, QFileDialog, QMessageBox

from ..product.mode import Frequency, Voltage, Current, CurveState
from ..screenshot import ScreenshotFileManager
from ..vision.detection import dump_elements, load_element, load_dict
import datetime
from ..element import Element
from ..element import Pin
from ..oscilloscope.oscilloscopemeasurer import compare_ivc_ex
from ..audio.wavplayer import WavPlayer, WavFile
# from ..audio.new_wav_player import WavPlayer
from time import time


__all__ = ["MainWindow"]


class SoundMode(enum.IntEnum):
    TEST_SOUND_MODE = 0
    COMPARE_SOUND_MODE = 1
    DEFAULT_SOUND_MODE = 2


class WithoutSound:

    def __init__(self, mainwin):
        self.mainwin = mainwin
        self.enable_sound_old = mainwin.enable_sound

    def __enter__(self):
        self.mainwin.set_enable_sound(False, notify_derived=False)

    def __exit__(self, *args):
        self.mainwin.enable_sound = self.enable_sound_old


class MainWindow(QMainWindow):
    """
    Base class for main window. Additionally manages trash
    with work modes and which of them are available.
    """

    score_changed = QtCore.pyqtSignal(float)
    score_threshold_changed = QtCore.pyqtSignal(float)

    enable_sound_support_changed = QtCore.pyqtSignal(bool)
    enable_sound_changed = QtCore.pyqtSignal(bool)

    frequency_changed = QtCore.pyqtSignal(Frequency)
    frequency_enabled_changed = QtCore.pyqtSignal(set)

    voltage_changed = QtCore.pyqtSignal(Voltage)
    voltage_enabled_changed = QtCore.pyqtSignal(set)

    current_changed = QtCore.pyqtSignal(Current)
    current_enabled_changed = QtCore.pyqtSignal(set)

    reference_curve_state_changed = QtCore.pyqtSignal(CurveState)
    test_curve_state_changed = QtCore.pyqtSignal(CurveState)
    external_freeze_state_changed = QtCore.pyqtSignal(CurveState)

    screenshots_allowed_changed = QtCore.pyqtSignal(bool)

    debug_changed = QtCore.pyqtSignal(bool)
    fps_changed = QtCore.pyqtSignal(float)

    tab_changed = QtCore.pyqtSignal(name="tab_changed")

    auto_calibration = QtCore.pyqtSignal()

    def __init__(
            self,
            screenshot_file_manager: ScreenshotFileManager,
            parent=None,
            *args,

            score: float = 0.0,
            score_threshold: float = 0.0,
            enable_score: bool = True,

            enable_sound_support: bool = False,
            enable_sound: bool = False,

            frequency_enabled: Optional[Set[Frequency]] = None,
            frequency: Frequency = Frequency.FREQUENCY_10HZ,

            voltage_enabled: Optional[Set[Voltage]] = None,
            voltage: Voltage = Voltage.VOLTAGE_3_3V,

            current_enabled: Optional[Set[Current]] = None,
            current: Current = Current.CURRENT_250UA,

            reference_curve_state: CurveState = CurveState.NORMAL,
            test_curve_state: CurveState = CurveState.NORMAL,
            external_freeze_state: CurveState = CurveState.NORMAL,

            screenshots_allowed: bool = True,
            screenshots_show_save_as_dialog: bool = True,

            debug: bool = False,
            fps: float = 0.0
    ) -> None:
        super().__init__(parent=parent)

        if len(args) != 0:
            raise ValueError("Positional arguments not supported")

        super().__init__(parent=parent)

        self.__screenshot_file_manager = screenshot_file_manager

        self.__tab_name = "S"
        self.__test_tabs = ["ZP", "TP"]  # Tabs for working with test plans
        self.__compare_tabs = ["C", "S"]  # Tabs for working in compare mode. "S" - not sure.
        dirname = os.path.dirname(__file__)
        filenameC = os.path.join(dirname, "..", "audio", "compare.wav")
        filenameT = os.path.join(dirname, "..", "audio", "test.wav")
        filenameS = []
        self.__compare_wav = WavFile(os.path.normpath(filenameC))
        self.__test_wav = WavFile(os.path.normpath(filenameT))
        self.__wav_last_time = 0.0
        self.__tab_curve_match = False
        self.wav_player = WavPlayer()
        self.wav_player.add_sound(filenameT, "test")
        self.wav_player.add_sound(filenameC, "compare")
        self.__score_wav = []
        for i in range(10):  # 10 files with sound with score
            filenameS.append(os.path.join(dirname, "..", "audio", str(i + 1) + ".wav"))
            self.__score_wav.append(WavFile(os.path.normpath(filenameS[i])))
            self.wav_player.add_sound(filenameS[i], str(i + 1))

        self.__score = score
        self.__score_threshold = score_threshold
        self.__enable_score = enable_score

        self.__enable_sound_support = enable_sound_support
        self.__enable_sound = enable_sound

        self.__frequency = frequency
        self.__frequency_prev = frequency
        self.__frequency_enabled = frequency_enabled\
            if frequency_enabled is not None else set(f for f in Frequency)
        self.__frequency_enabled_original = set(self.__frequency_enabled)

        self.__voltage = voltage
        self.__voltage_prev = voltage
        self.__voltage_enabled = voltage_enabled\
            if voltage_enabled is not None else set(v for v in Voltage)
        self.__voltage_enabled_original = set(self.__voltage_enabled)

        self.__current = current
        self.__current_prev = current
        self.__current_enabled = current_enabled\
            if current_enabled is not None else set(c for c in Current)
        self.__current_enabled_original = set(self.__current_enabled)

        self.__reference_curve_state = reference_curve_state
        self.__test_curve_state = test_curve_state
        self.__external_freeze_state = external_freeze_state
        self.__reference_curve_state_backup = self.__reference_curve_state
        self.__test_curve_state_backup = self.__test_curve_state

        self.__screenshots_allowed = screenshots_allowed
        self.__screenshots_show_save_as_dialog =\
            screenshots_show_save_as_dialog

        self.__debug = debug
        self.__fps = fps
        self.__reference_curve = None
        self.__test_curve = None
        self.__reference_curve_override = None
        self.__test_curve_override = None
        self.__devices_connected = False
        self.__file_not_found = False
        self.__devices_not_correct = False

        self.__test_plan_pin_index = 0

        self.tab_changed.connect(self.__on_tab_changed)

        self.__pins_dict = None
        self._no_flash_drive = False  # it is analog of self__.file_not_found but used for alert messages

        # TODO: make Enum Tab_Name
        self.__prev_params = {}
        for tab in ["S", "ZP", "TP", "C"]:
            self.__prev_params[tab] = {}
            self.__prev_params[tab]["frequency"] = self.__frequency
            self.__prev_params[tab]["voltage"] = self.__voltage
            self.__prev_params[tab]["current"] = self.__current

        self.scroller = self.Scroller(self.__test_plan_pin_refresh)
        self.__make_captions()
        self.__make_incorrect_device_warning_template()

    class Scroller:
        """
        # this class helps us to manage all QTimer stuff and scrolls pins (self.zp_push_button_right and other buttons)
        #
        # the signal emited by pressing the buttons on the panels tp and zp
        # connected to scroll_left and scroll_right functions
        # the signal emited by releasing this buttons connected to the stop
        """
        def __init__(self, scroll_func):
            self._scroll = scroll_func

            # timer config
            self._timer = QtCore.QTimer()               # Create timer
            self._timer.timeout.connect(self._on_time)  # Connect timer
            self._t = 0                                 # Time elapsed from press button
            self._period = 500                          # timer period i.e scroll period in milliseconds
            self._scroll_speed = 0.001                  # scroll parameter
            self._direction = "right"                   # scroll direction (> 0 - right; < 0 - left)
            self._stop_timer = True                     # Timer stop indicator

        def scroll_left(self):
            self._direction = "left"
            # if it the first press we need to count 1sec to start fast scrolling
            self._stop_timer = False
            self._timer.setInterval(1000)
            self._timer.start()
            self._scroll(-1)

        def scroll_right(self):
            self._direction = "right"
            # if it the first press we need to count 1sec to start fast scrolling
            self._stop_timer = False
            self._timer.setInterval(1000)
            self._timer.start()
            self._scroll(1)

        def stop(self):
            # when we release the button we need to reset time and set stop indicator
            self._t = 0
            self._timer.stop()
            self._stop_timer = True

        def _on_time(self):
            # check for timer stop
            if not self._stop_timer:
                self._timer.setInterval(self._period)
                self._timer.start()

            self._t += self._period
            try:
                # Calculate diff to the next point
                index_diff = round(math.exp(self._scroll_speed * self._t))
            except OverflowError:
                index_diff = 0
            if self._direction == "left":
                # if left - we need to change sign
                index_diff *= -1
            # make scroll on calculated step
            self._scroll(index_diff)

    def __on_tab_changed(self):
        self.__test_plan_store_prev("frequency", self.__frequency)
        self.__test_plan_store_prev("voltage", self.__voltage)
        self.__test_plan_store_prev("current", self.__current)

    def on_flash_disconnect(self):
        # we don't need to change anything on S, C tab
        if self.__tab_name in ["S", "C"]:
            return
        # and here we just need to refresh main window
        self.__file_not_found = True  # I'm not sure that this is so necessary here to set False, but for confidence
        self.__refresh_main_widget()

    def on_flash_connected(self):
        """
        this function is called from application.py
        on __screenshot_file_manager.external_media_state_changed signal
        be careful - at start this function is called twice
        """
        self.__test_plan_pin_index = 0  # roll back to the first point on ZP and TP tabs

        # we should reload __pins_dict from flash drive
        self.__pins_dict = None

        # we should refresh the tab window (it works the same as tp or zp button push)
        if self.__tab_name == "TP":
            self.on_test_plan_tab_tp()
        if self.__tab_name == "ZP":
            self.on_test_plan_tab_zp()
        # change tha state
        self.__file_not_found = False
        self.__refresh_main_widget()

    def __refresh_main_widget(self):
        visible = self.ivc_viewer
        if not self.__devices_connected and not self.__devices_not_correct:
            visible = self.label_DeviceStatus
        elif self.__file_not_found:
            visible = self.label_FileNotFound
        elif self.__devices_not_correct:
            visible = self.label_DevNotCorrect
        for wdg in (
                self.ivc_viewer,
                self.label_DevNotCorrect,
                self.label_DeviceStatus,
                self.label_FileNotFound):
            wdg.setVisible(wdg is visible)

    def set_devices_ready(self, connected):
        self.__devices_connected = connected
        self.__refresh_main_widget()

    def __make_incorrect_device_warning_template(self):
        text = ""
        t = QtCore.QCoreApplication.translate("t", "Версия библиотеки ")
        text = text + t
        text = text + "{}: {}\n"
        t = QtCore.QCoreApplication.translate("t", "\nВерсия прошивки: ")
        text = text + t
        text = text + "{}\n"
        t = QtCore.QCoreApplication.translate("t", "\nЭта версия библиотеки "
                                                   "не подходит для данной прошивки.\n "
                                                   "Версии прошивок, которые совместимы с "
                                                   "данной библиотекой:\n"
                                              )
        text = text + t
        text = text + "{}"
        t = QtCore.QCoreApplication.translate("t", "За помощью обращайтесь в техническую поддержку\n "
                                                   "ЦИФ МГУ по email адресу:  info@ximc.ru")
        text += t
        self.__incorrect_device_warning_template = text

    def set_devices_correct(self, correct, name, version, firmw, firms):
        self.label_DevNotCorrect.setStyleSheet("color: rgb(255, 0, 0); font: bold 14px;")
        self.label_DevNotCorrect.setAlignment(QtCore.Qt.AlignCenter)
        self.__devices_not_correct = correct

        # Make compatible firmware list (text)
        compatible_fw_list = "\n"
        for f in firms:
            compatible_fw_list += str(f)
            compatible_fw_list += "\n"
        # Make message
        test_text = self.__incorrect_device_warning_template.format(name, version, firmw, compatible_fw_list)
        self.label_DevNotCorrect.setText(test_text)
        self.__refresh_main_widget()

    def __play_match_wav(self, score=0.0):
        if not self.enable_sound:
            return
        match = self.__score < self.__score_threshold
        if self.get_sound_mode() == SoundMode.COMPARE_SOUND_MODE and not match:
            for i in range(len(self.__score_wav)):
                if (score < (i + 1) / 10) and score > (i / 10):
                    self.wav_player.play(str(i + 1))
                else:
                    pass

        elif match != self.__tab_curve_match:
            self.__tab_curve_match = match
            if match and self.get_sound_mode() == SoundMode.TEST_SOUND_MODE:
                now = time()
                diff = now - self.__wav_last_time
                if diff >= self.__test_wav.duration:
                    self.__wav_last_time = now
                    self.wav_player.play("test")

    def _set_curves(
            self,
            reference_curve, test_curve,
            frequency: Frequency, voltage: Voltage, current: Current
    ) -> None:
        pass

    def set_curves(
            self,
            reference_curve, test_curve,
            frequency: Frequency, voltage: Voltage, current: Current
    ) -> None:
        self.__reference_curve = reference_curve
        self.__test_curve = test_curve
        if self.__reference_curve_override is not None:
            reference_curve = self.__reference_curve_override
        if self.__test_curve_override is not None:
            test_curve = self.__test_curve_override
        self._set_curves(
            reference_curve,
            test_curve,
            frequency,
            voltage,
            current
        )
        if self.__enable_score:
            self.__score = compare_ivc_ex(
                reference_curve,
                test_curve,
                voltage,
                current
            )
            self.__play_match_wav(self.__score)
            self._set_score(self.__score)
            self.score_changed.emit(self.__score)

    def _set_score(self, score: float) -> None:
        pass

    @property
    def score(self):
        return self.__score

    def _set_score_threshold(self, score_threshold: float) -> None:
        pass

    def __refresh_score_threshold_gui(self):
        value = self.score_threshold
        if value < 0.0:
            value = 0.0
        if value > 1.0:
            value = 1.0
        self.pushButton_score_threshold_minus.setEnabled(value > 0.0)
        self.pushButton_score_threshold_plus.setEnabled(value < 1.0)
        self.label_score_threshold_value.setText(
            "{value:s}".format(value=str(int(self.score_threshold * 100)) + "%")
        )

    def on_score_threshold_button(self, add):
        if add:
            self.score_threshold = min(1.0, self.score_threshold + 0.05)
        else:
            self.score_threshold = max(0.0, self.score_threshold - 0.05)
        self.__refresh_score_threshold_gui()

    def get_score_threshold(self) -> float:
        return self.__score_threshold

    @QtCore.pyqtSlot(float)
    def set_score_threshold(self, score_threshold: float) -> None:
        self.__score_threshold = score_threshold
        self._set_score_threshold(self.__score_threshold)
        self.score_threshold_changed.emit(self.__score_threshold)

    score_threshold = property(get_score_threshold, set_score_threshold)

    def _set_enable_sound_support(self, enable_sound_support: bool) -> None:
        pass

    def get_enable_sound_support(self) -> bool:
        return self.__enable_sound_support

    @QtCore.pyqtSlot(bool)
    def set_enable_sound_support(self, enable_sound_support: bool) -> None:
        self.__enable_sound_support = enable_sound_support
        self._set_enable_sound_support(self.__enable_sound_support)
        self.enable_sound_support_changed.emit(self.__enable_sound_support)

    enable_sound_support = property(
        get_enable_sound_support, set_enable_sound_support)

    def _set_enable_sound(self, enable_sound: bool) -> None:
        pass

    def get_enable_sound(self) -> bool:
        return self.__enable_sound

    @QtCore.pyqtSlot(bool)
    def set_enable_sound(self, enable_sound: bool, notify_derived=True) -> None:
        self.__enable_sound = enable_sound
        if notify_derived:
            self._set_enable_sound(self.__enable_sound)
        self.enable_sound_changed.emit(self.__enable_sound)

    enable_sound = property(get_enable_sound, set_enable_sound)

    # we have 4 tabs
    # 1 - IVC comapre mode - 'S'
    # 2 - Test plan creation - 'ZP'
    # 3 - Plan testing - 'TP'
    # 4 - EP1 configuration - 'C'
    #  __tab_name - name of tab in UI

    def get_sound_mode(self) -> SoundMode:
        if self.__tab_name in self.__test_tabs:
            return SoundMode.TEST_SOUND_MODE
        if self.__tab_name in self.__compare_tabs:
            return SoundMode.COMPARE_SOUND_MODE
        return SoundMode.DEFAULT_SOUND_MODE

    def _set_frequency(self, frequency: Frequency) -> None:
        pass

    def get_frequency(self) -> Frequency:
        return self.__frequency

    @QtCore.pyqtSlot(Frequency)
    def set_frequency(self, frequency: Frequency) -> None:
        self.__test_plan_store_prev("frequency", self.__frequency)
        self.__frequency = frequency
        self._set_frequency(self.__frequency)
        self.frequency_changed.emit(self.__frequency)

    frequency = property(get_frequency, set_frequency)

    def _set_frequency_enabled(
            self, frequency_enabled: Optional[Set[Frequency]]) -> None:
        pass

    def get_frequency_enabled(self) -> Set[Frequency]:
        return self.__frequency_enabled

    @QtCore.pyqtSlot(Set[Frequency])
    def set_frequency_enabled(
            self, frequency_enabled: Optional[Set[Frequency]]) -> None:
        self.__frequency_enabled = frequency_enabled\
            if frequency_enabled is not None else set(f for f in Frequency)
        self._set_frequency_enabled(self.__frequency_enabled)
        self.frequency_enabled_changed.emit(self.__frequency_enabled)

    @QtCore.pyqtSlot()
    def clear_frequency_enabled(self):
        self.__frequency_enabled = set(f for f in Frequency)

    frequency_enabled = property(
        get_frequency_enabled,
        set_frequency_enabled,
        clear_frequency_enabled
    )

    def _set_voltage(self, voltage: Voltage) -> None:
        pass

    def get_voltage(self) -> Voltage:
        return self.__voltage

    @QtCore.pyqtSlot(Voltage)
    def set_voltage(self, voltage: Voltage) -> None:
        self.__test_plan_store_prev("voltage", self.__voltage)
        self.__voltage = voltage
        self._set_voltage(self.__voltage)
        self.voltage_changed.emit(self.__voltage)

    voltage = property(get_voltage, set_voltage)

    def _set_voltage_enabled(
            self, voltage_enabled: Optional[Set[Voltage]]) -> None:
        pass

    def get_voltage_enabled(self) -> Set[Voltage]:
        return self.__voltage_enabled

    @QtCore.pyqtSlot(Set[Voltage])
    def set_voltage_enabled(
            self, voltage_enabled: Optional[Set[Voltage]]) -> None:
        self.__voltage_enabled = voltage_enabled\
            if voltage_enabled is not None else set(f for f in Voltage)
        self._set_voltage_enabled(self.__voltage_enabled)
        self.voltage_enabled_changed.emit(self.__voltage_enabled)

    @QtCore.pyqtSlot()
    def clear_voltage_enabled(self):
        self.__voltage_enabled = set(f for f in Voltage)

    voltage_enabled = property(
        get_voltage_enabled, set_voltage_enabled, clear_voltage_enabled)

    def _set_current(self, current: Current) -> None:
        pass

    def get_current(self) -> Current:
        return self.__current

    @QtCore.pyqtSlot(Current)
    def set_current(self, current: Current) -> None:
        self.__test_plan_store_prev("current", self.__current)
        self.__current = current
        self._set_current(self.__current)
        self.current_changed.emit(self.__current)

    current = property(get_current, set_current)

    def _set_current_enabled(
            self, current_enabled: Optional[Set[Current]]) -> None:
        pass

    def get_current_enabled(self) -> Set[Current]:
        return self.__current_enabled

    @QtCore.pyqtSlot(Set[Current])
    def set_current_enabled(
            self, current_enabled: Optional[Set[Current]]) -> None:
        self.__current_enabled = current_enabled\
            if current_enabled is not None else set(f for f in Current)
        self._set_current_enabled(self.__current_enabled)
        self.current_enabled_changed.emit(self.__current_enabled)

    @QtCore.pyqtSlot()
    def clear_current_enabled(self):
        self.__current_enabled = set(f for f in Current)

    current_enabled = property(
        get_current_enabled, set_current_enabled, clear_current_enabled)

    def _set_reference_curve_state(
            self, reference_curve_state: CurveState) -> None:
        pass

    def get_reference_curve_state(self) -> CurveState:
        return self.__reference_curve_state

    @QtCore.pyqtSlot(CurveState)
    def set_reference_curve_state(
            self, reference_curve_state: CurveState) -> None:
        if self.__external_freeze_state == CurveState.NORMAL:
            self.__reference_curve_state = reference_curve_state
            self._set_reference_curve_state(self.__reference_curve_state)
            self.reference_curve_state_changed.emit(
                self.__reference_curve_state)

    reference_curve_state = property(
        get_reference_curve_state, set_reference_curve_state)

    def _set_test_curve_state(self, test_curve_state: CurveState) -> None:
        pass

    def get_test_curve_state(self) -> CurveState:
        return self.__test_curve_state

    @QtCore.pyqtSlot(CurveState)
    def set_test_curve_state(self, test_curve_state: CurveState) -> None:
        if self.__external_freeze_state == CurveState.NORMAL:
            self.__test_curve_state = test_curve_state
            self._set_test_curve_state(self.__test_curve_state)
            self.test_curve_state_changed.emit(self.__test_curve_state)

    test_curve_state = property(get_test_curve_state, set_test_curve_state)

    def _set_external_freeze_state(
            self, external_freeze_state: CurveState) -> None:
        pass

    def get_external_freeze_state(self) -> CurveState:
        return self.__external_freeze_state

    @QtCore.pyqtSlot(CurveState)
    def set_external_freeze_state(
            self, external_freeze_state: CurveState) -> None:
        if self.__external_freeze_state != external_freeze_state:
            self.__external_freeze_state = external_freeze_state
            if self.__external_freeze_state == CurveState.NORMAL:
                self.__reference_curve_state =\
                    self.__reference_curve_state_backup
                self.__test_curve_state = self.__test_curve_state_backup
            else:
                self.__reference_curve_state_backup =\
                    self.__reference_curve_state
                self.__reference_curve_state = CurveState.FROZEN
                self.__test_curve_state_backup = self.__test_curve_state
                self.__test_curve_state = CurveState.FROZEN

            self._set_external_freeze_state(self.__external_freeze_state)
            self.external_freeze_state_changed.emit(
                self.__external_freeze_state)

    external_freeze_state = property(
        get_external_freeze_state, set_external_freeze_state)

    def _set_screenshots_allowed(self, screenshots_allowed: bool) -> None:
        pass

    def get_screenshots_allowed(self) -> bool:
        return self.__screenshots_allowed

    @QtCore.pyqtSlot(bool)
    def set_screenshots_allowed(self, screenshots_allowed: bool) -> None:
        self.__screenshots_allowed = screenshots_allowed
        self._set_screenshots_allowed(screenshots_allowed)
        self.screenshots_allowed_changed.emit(self.__screenshots_allowed)

    screenshots_allowed = property(
        get_screenshots_allowed, set_screenshots_allowed)

    def _set_debug(self, debug: bool) -> None:
        pass

    def get_debug(self) -> bool:
        return self.__debug

    @QtCore.pyqtSlot(bool)
    def set_debug(self, debug: bool) -> None:
        self.__debug = debug
        self._set_debug(self.__debug)
        self.debug_changed.emit(self.__debug)

    debug = property(get_debug, set_debug)

    def _set_fps(self, fps: float) -> None:
        pass

    def get_fps(self) -> float:
        return self.__fps

    @QtCore.pyqtSlot(float)
    def set_fps(self, fps: float) -> None:
        self.__fps = fps
        self._set_fps(self.__fps)
        self.fps_changed.emit(self.__fps)

    fps = property(get_fps, set_fps)

    def _make_screenshot(self) -> Optional[QPixmap]:
        pass

    def __make_captions(self):
        self.__screen_caption = QtCore.QCoreApplication.translate("t", "Сохранение изображения")
        self.__board_caption = QtCore.QCoreApplication.translate("t", "Сохранение ВАХ")

    @QtCore.pyqtSlot()
    def make_screenshot(self) -> None:
        with WithoutSound(self):
            screenshot = self._make_screenshot()
            if screenshot is None:
                return

            screenshot_path = self.__screenshot_file_manager\
                .get_next_screenshot_file_name()
            if screenshot_path is None and not self.__screenshots_show_save_as_dialog:
                return

            if self.__screenshots_show_save_as_dialog:
                if screenshot_path is None:
                    date = datetime.datetime.now()
                    screenshot_path = os.path.join("./", "ivc_" + date.strftime("%Y-%m-%d-%H.%M.%S"))
                screenshot_path = QFileDialog.getSaveFileName(
                    parent=self,
                    caption=self.__screen_caption,
                    directory=screenshot_path,
                    filter="Images (*.png)"
                )[0]
                if len(screenshot_path) == 0:
                    return
            screenshot.save(screenshot_path)
            self.__screenshot_file_manager.commit_screenshot(screenshot_path)
            print("Screenshot saved to: \"{}\".".format(screenshot_path))

    def on_freeze_gpio(self, high):
        with WithoutSound(self):
            tab_index = self.test_plan_tab_widget.currentIndex()
            if tab_index == 0:
                self.set_external_freeze_state(
                    CurveState.FROZEN if high else CurveState.NORMAL)
            elif tab_index == 1:
                if high:
                    self.on_test_plan_zp_save()
            elif tab_index == 2:
                if high:
                    self.__test_plan_pin_refresh(1)

    def __get_path_and_tp_elem(self, create_new_file=None) -> Optional[Tuple[str, Element]]:
        """
        this function check existence of path to board.json file
        if there is no such file it tries to create this path
        this function just create path, not board json file ( file creates when we want to add points or on tp/zp tab)
        in case of failure it returns one None
        in case of success it try to read, or if need reload pins dict
        and return path and Element
        Element can be none if there is no board.json file
        or if there non valid json
        """
        with WithoutSound(self):
            self.__file_not_found = True
            reference_path = self.__screenshot_file_manager.get_reference_path()
            if (self.__screenshots_show_save_as_dialog and not reference_path) or create_new_file:
                reference_path = QFileDialog.getSaveFileName(
                    parent=self,
                    caption=self.__board_caption,
                    directory=os.path.join("./", "board"),
                    filter="JSON file (*.json)"
                )[0] or reference_path
                self.__screenshot_file_manager.last_reference_path = reference_path
            if reference_path is None:
                return None
            if not self.__screenshots_show_save_as_dialog:
                try:
                    if not os.path.isdir(reference_path):
                        os.makedirs(reference_path, exist_ok=True)
                except Exception as ex:
                    print("Failed to make dirs: {}.".format(reference_path))
                    print(ex)
                    return None

                board_path = os.path.join(reference_path, "board.json")
            else:
                board_path = reference_path
            if os.path.isfile(board_path):
                print("DEBUG:\t thereis no board json")
            try:
                if self.__pins_dict is None:
                    self.__pins_dict = load_dict(board_path)

                element = load_element(self.__pins_dict)
                self.__file_not_found = False
            except FileNotFoundError:
                print("there is no file board.json ")
                element = None
            except Exception:
                element = None
                print("board.json parsing failed")
            return board_path, element

    def on_test_plan_zp_save_new_file(self):
        with WithoutSound(self):
            pin = self.__make_pin(default_values=False)
            if pin is None:
                return
            path_elem = self.__get_path_and_tp_elem(create_new_file=True)
            if path_elem is None:
                self.__file_not_found = True
                return
            board_path, element = path_elem
            if element is not None:
                element.pins[self.__test_plan_pin_index] = pin
                dump_elements(board_path, [element])
                self.__pins_dict = load_dict(board_path)
            self.__test_plan_pin_refresh(0)

    def on_test_plan_zp_right(self):
        self.__test_plan_pin_refresh(1)

    def on_test_plan_zp_left(self):
        self.__test_plan_pin_refresh(-1)

    def on_test_plan_tp_right(self):
        self.__test_plan_pin_refresh(1)

    def on_test_plan_tp_left(self):
        self.__test_plan_pin_refresh(-1)

    def __test_plan_pin_refresh(self, index_diff=0):
        with WithoutSound(self):
            path_elem = self.__get_path_and_tp_elem()
            if path_elem is None:
                return
            board_path, element = path_elem
            if element is None or len(element.pins) == 0:
                self.__file_not_found = True
                return
            self.__file_not_found = False
            pins = element.pins
            self.__test_plan_pin_index += index_diff
            if self.__test_plan_pin_index < 0:
                self.__test_plan_pin_index = 0
            if self.__test_plan_pin_index >= len(element.pins):
                self.__test_plan_pin_index = len(element.pins) - 1
            pin = element.pins[self.__test_plan_pin_index]
            self.__test_curve_override = pin.ivc
            self.__refresh_main_widget()
            self.test_plan_ui_refresh(
                index=self.__test_plan_pin_index,
                max_num=len(pins),
                left=(element is not None) and (self.__test_plan_pin_index > 0),
                right=(element is not None) and (self.__test_plan_pin_index < (len(pins) - 1)),
                file_found=not self.__file_not_found
            )
            # if we get last pin by long press: i.e with Scroller,
            # we have to stop timer because there is no signal emited by button release
            # because in test_plan_ui_refresh button setEnabled(False) instead of release
            if (self.__test_plan_pin_index <= 0) or (self.__test_plan_pin_index >= len(pins) - 1):
                self.scroller.stop()
            self.__test_plan_store_prev("frequency", self.__frequency)
            self.__test_plan_store_prev("voltage", self.__voltage)
            self.__test_plan_store_prev("current", self.__current)
            self.frequency = pin.frequency
            self.voltage = pin.voltage
            self.current = pin.current

    def __test_plan_restore(self):
        self.frequency = self.__prev_params[self.__tab_name]["frequency"]
        self.voltage = self.__prev_params[self.__tab_name]["voltage"]
        self.current = self.__prev_params[self.__tab_name]["current"]

    def __test_plan_store_prev(self, param, prev_value):
        try:
            self.__prev_params[self.__tab_name][param] = prev_value
        except KeyError:
            pass

    def __switch_tab(self, name):
        self.tab_changed.emit()
        self.__tab_name = name
        self.__tab_curve_match = False

    def on_test_plan_tab_s(self):
        self.__switch_tab("S")
        self.__test_plan_restore()
        self.set_frequency_enabled(self.__frequency_enabled_original)
        self.set_voltage_enabled(self.__voltage_enabled_original)
        self.set_current_enabled(self.__current_enabled_original)
        self.__file_not_found = False
        self.__reference_curve_override = None
        self.__test_curve_override = None

    def __test_plan_create_pin(self, only_if_empty):
        print("DEBUG:\tSTART __test_plan_create_pin")
        with WithoutSound(self):
            path_elem = self.__get_path_and_tp_elem()
            if path_elem is None:
                self.__file_not_found = True
                return
            board_path, element = path_elem
            do_save = False
            if element is None:
                # if there is no file board.json we don't need to rise alert
                # we just need to create default file
                if os.path.isfile(board_path):
                    print("DEBUG:\twe have board file, but it is not valid")
                    self._no_flash_drive = False
                    msg = QMessageBox()
                    msg.setIcon(QMessageBox.Warning)
                    msg.setText("Файл плана тестирования не соответствует требуемому формату.\n"
                                "Нажмите ОК, чтобы удалить текущий файл и создать новый.\n"
                                "Чтобы сохранить текущую версию файла, извлеките накопитель и воспользуйтесь ПК.")
                    msg.setWindowTitle("Eyepoint S1")
                    msg.setStandardButtons(QMessageBox.Ok)
                    msg.exec_()

                    # check that flash drive is plugged in
                    test = self.__get_path_and_tp_elem()
                    if test is None:
                        # if user pulled out a flash drive we return from function, that creates pin
                        return
                    # else user want to delete bad json

                element = Element(
                    pins=[],
                    is_manual=True,
                    center=np.zeros(2)
                )
                do_save = True
            if not only_if_empty or len(element.pins) == 0:
                pin = self.__make_pin(default_values=True)
                if pin is None:
                    return
                element.pins.append(pin)
                self.__test_plan_pin_index = len(element.pins) - 1
                self.frequency = pin.frequency
                self.voltage = pin.voltage
                self.current = pin.current
                do_save = True
            if do_save:
                dump_elements(board_path, [element])
                self.__pins_dict = load_dict(board_path)
            self.__test_plan_pin_refresh()
            print("DEBUG:\tEND __test_plan_create_pin")

    def __make_pin(self, default_values):
        with WithoutSound(self):
            if self.__reference_curve is None:
                return None

            if default_values:
                ms = self.__reference_curve[0]
                ivc = (
                    ms,
                    (
                        # Sin probe signal
                        np.sin(np.linspace(0, 2 * np.pi, ms.n_points)),
                        np.zeros(ms.n_points)
                    )
                )
            else:
                ivc = self.__reference_curve
            return Pin(
                rc=np.zeros(2),
                ivc=ivc,
                reference_ivc=None,
                frequency=self.frequency,
                voltage=self.voltage,
                current=self.current
            )

    def on_test_plan_tab_zp(self):
        self.__switch_tab("ZP")
        self.reference_curve_state = CurveState.NORMAL
        self.test_curve_state = CurveState.NORMAL
        self.set_frequency_enabled(self.__frequency_enabled_original)
        self.set_voltage_enabled(self.__voltage_enabled_original)
        self.set_current_enabled(self.__current_enabled_original)
        self.__test_plan_restore()
        self.__test_plan_create_pin(only_if_empty=True)

    def on_test_plan_zp_save(self):
        with WithoutSound(self):
            pin = self.__make_pin(default_values=False)
            if pin is None:
                return
            path_elem = self.__get_path_and_tp_elem()
            if path_elem is None:
                self.__file_not_found = True
                return
            board_path, element = path_elem
            if element is not None:
                element.pins[self.__test_plan_pin_index] = pin
                dump_elements(board_path, [element])
                self.__pins_dict = load_dict(board_path)
            self.__test_plan_pin_refresh(0)

    def on_test_plan_zp_new_point(self):
        self.__test_plan_create_pin(only_if_empty=False)

    def on_test_plan_tab_tp(self):
        with WithoutSound(self):
            self.__switch_tab("TP")
            self.reference_curve_state = CurveState.NORMAL
            self.test_curve_state = CurveState.NORMAL
            self.__test_plan_pin_refresh()
            self.set_frequency_enabled(set())
            self.set_voltage_enabled(set())
            self.set_current_enabled(set())

    def on_test_plan_tp_save(self):
        self.make_screenshot()

    def on_test_plan_tab_c(self):
        with WithoutSound(self):
            self.__switch_tab("C")
            self.__test_plan_restore()
            self.set_frequency_enabled(self.__frequency_enabled_original)
            self.set_voltage_enabled(self.__voltage_enabled_original)
            self.set_current_enabled(self.__current_enabled_original)
            self.__file_not_found = False
            self.__reference_curve_override = None
            self.__test_curve_override = None
            self.__refresh_score_threshold_gui()

    def on_test_plan_auto_calibration(self):
        with WithoutSound(self):
            self.auto_calibration.emit()
